import unittest.mock as mock

import pop.hub
import pytest
import yaml

# Things left to test:
# Full ordered tests, with all attributes on multiple layers
# CLI arg Groups


def test_source():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load(["m1", "m2"], "m1")
    assert hub.OPT.m2.base == "single"


def test_load():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load("c1", "c1")
    assert hub.OPT.c1.int == 2016
    assert hub.OPT.c1.string == "cheese"
    assert hub.OPT.c1.test == 47


def test_render_noargs():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load("rend1", "rend1")
    assert hub.OPT.rend1.yaml == {"rubber": "Svien"}
    assert hub.OPT.rend1.json == {"rubber": "Svien"}


def test_render_args():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch(
        "sys.argv",
        [
            "root",
            "--yaml",
            "test: True",
            "--json",
            '{"test": true}',
            "--cli",
            "test=True",
        ],
    ):
        hub.config.integrate.load("rend1", "rend1")
    assert hub.OPT.rend1.yaml == {"test": True}
    assert hub.OPT.rend1.json == {"test": True}
    assert hub.OPT.rend1.cli == ({"test": True},)


def test_os():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("os.environ", {"FOOTEST": "osvar", "POPTESTOSVAR": "good"}):
        hub.config.integrate.load("c1", "c1")
    # Lowercase works
    assert hub.OPT.c1.foo == "osvar"
    # Standard works with cli enabled but not called
    assert hub.OPT.c1.osvar == "good"
    # No OSVAR passed
    assert hub.OPT.c1.lastos == "no os"


def test_os_config(tmp_path):
    """
    test config as an os environment variable
    """
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    conf_file = tmp_path / "test"

    with open(conf_file, "w") as fp:
        yaml.safe_dump({"c1": {"config_test": "yes"}}, fp)

    with mock.patch("os.environ", {"C1_CONFIG": str(conf_file)}):
        hub.config.integrate.load("c1", "c1")

    assert hub.OPT.c1.config_test == "yes"


def test_log():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load("c1", "c1")
    assert hub.OPT.c1.log_file


def test_args():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "--test", "new"]):
        hub.config.integrate.load("c1", "c1")
    assert hub.OPT.c1.test == "new"


def test_no_args():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "--foo", "new"]):
        with pytest.raises(SystemExit):
            hub.config.integrate.load("c1", "c1")


def test_actions():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "-A"]):
        hub.config.integrate.load("c1", "c1")
    assert hub.OPT.c1.act is True


def test_cli_over_os():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("os.environ", {"FOOTEST": "osvar", "POPTESTOSVAR": "good"}):
        with mock.patch("sys.argv", ["config", "--osvar", "new"]):
            hub.config.integrate.load("c1", "c1")

    # Lowercase works
    assert hub.OPT.c1.foo == "osvar"
    # Standard works
    assert hub.OPT.c1.osvar == "new"
    # No OSVAR passed
    assert hub.OPT.c1.lastos == "no os"


def test_dyne_cli():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch(
        "sys.argv",
        [
            "root",
            "--dyned1",
            "DYNED1",
            "--exdyned1",
            "EXDYNED1",
            "--exdyned2",
            "EXDYNED2",
            "--exdyned3",
            "EXDYNED3",
            "--exdyned4",
            "EXDYNED4",
            "ext_test",
        ],
    ):
        hub.config.integrate.load("d1", "d1")
    assert hub.OPT.d1.dyned1 == "DYNED1"
    assert hub.OPT.d1.exdyned1 == "EXDYNED1"
    assert hub.OPT.d1.exdyned2 == "EXDYNED2"
    assert hub.OPT.d1.exdyned3 == "EXDYNED3"
    assert hub.OPT.d1.exdyned4 == "EXDYNED4"


def test_dyne_defaults():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    hub.config.integrate.load("d1", "d1")
    assert hub.OPT.d1.dyned1 == "default"


def test_subcommands_single():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch(
        "sys.argv",
        [
            "root",
            "test",
            "--single",
            "another!",
        ],
    ):
        hub.config.integrate.load("s1", "s1")
    # assert hub.SUBPARSER == "test"
    assert hub.OPT.s1.single == "another!"


def test_subcommands_double_1():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch(
        "sys.argv",
        [
            "root",
            "test",
            "--double",
            "another!",
        ],
    ):
        hub.config.integrate.load("s1", "s1")
    assert hub.OPT.s1.double == "another!"


def test_subcommands_double_2():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "valid", "--single", "another!"]):
        with pytest.raises(SystemExit):
            hub.config.integrate.load("s1", "s1")


def test_subcommands_invalid():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch(
        "sys.argv",
        [
            "root",
            "test",
            "--double",
            "another!",
        ],
    ):
        hub.config.integrate.load("s1", "s1")
    assert hub.OPT.s1.double == "another!"


def test_positional():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "foo", "bar"]):
        hub.config.integrate.load("p1", "p1")
    assert hub.OPT.p1.first == "foo"
    assert hub.OPT.p1.second == "bar"


def test_version():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    with mock.patch("sys.argv", ["config", "--version"]):
        with pytest.raises(SystemExit):
            hub.config.integrate.load("c1", "c1")
