import asyncio
import logging
import os
import pathlib
import socket
import tempfile
import unittest.mock as mock
from typing import Any
from typing import Union

import pop.hub
import pytest

LOGGER = logging.getLogger()


@pytest.fixture
def hub():
    hub = pop.hub.Hub()
    hub.pop.sub.add(dyne_name="config")
    yield hub


@pytest.fixture(name="logfile")
def logfile_():
    with tempfile.NamedTemporaryFile(suffix=".log") as fh:
        yield fh


@pytest.fixture(name="dgram")
def datagram_socket():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("127.0.0.1", 9000))
    yield sock
    sock.close()


@pytest.fixture(name="unix")
def unix_socket(logfile):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(logfile.name)
    yield sock
    sock.close()


@pytest.fixture(name="sock")
def socket_(logfile):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 9000))
    yield sock
    sock.close()


def test_basic(hub, logfile):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=basic",
            "--log-level=trace",
            f"--log-file={logfile.name}",
            "--log-handler-options",
            "mode=a",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    contents = logfile.read()
    assert b"[TRACE   ] hello world\n" in contents


@pytest.mark.skip(reason="Test needs debugging")
def test_datagram_dgram(hub, dgram):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=datagram",
            "--log-level=trace",
            f"--log-file=127.0.0.1:9000",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    chunk = dgram.recv(1024)

    assert b"[TRACE   ] hello world\n" in chunk


@pytest.mark.skip(reason="Test needs debugging")
def test_datagram_unix(hub, logfile, unix):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=datagram",
            "--log-level=trace",
            f"--log-file={logfile.name}",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    chunk = unix.recv(1024)

    assert b"[TRACE   ] hello world\n" in chunk


@pytest.mark.skip(reason="Test needs debugging")
def test_socket(hub, sock):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=socket",
            "--log-level=trace",
            f"--log-file=127.0.0.1:9000",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    chunk = sock.recv(1024)

    assert b"[TRACE   ] hello world\n" in chunk


@pytest.mark.skipif(not os.name == "nt", reason="Test only runs on windows")
@pytest.mark.skip(reason="Test needs debugging")
def test_nt_event(hub):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=nt_event",
            "--log-level=trace",
            "--log-handler-options",
            "appname=my_app",
            "logtype=Application",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    chunk = "TODO get nt_event"

    assert b"[TRACE   ] hello world\n" in chunk


def test_null(hub, caplog):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=null",
            "--log-level=trace",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("[hello world]")

    assert "[hello world]\n" in caplog.text


def test_rotating(hub, logfile):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=rotating",
            "--log-level=trace",
            f"--log-file={logfile.name}",
            "--log-handler-options",
            "mode=a",
            "maxBytes=1",
            "backupCount=1",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    # Make sure we overflow and roll the log
    for _ in range(10):
        hub.log.trace("hello world")

    with open(f"{logfile.name}.1") as fh:
        contents = fh.read()
    assert "[TRACE   ] hello world\n" in contents


def test_timed_rotating(hub, logfile):
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=timed_rotating",
            "--log-level=trace",
            f"--log-file={logfile.name}",
            "--log-handler-options",
            "when=S",
            "interval=0.5",
            "backupCount=2",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")
    hub.log.trace("hello there")

    lf = pathlib.Path(logfile.name).parent
    rotated_log = next(lf.glob("*.log.*"))

    with open(rotated_log) as fh:
        contents = fh.read()

    assert "[TRACE   ] hello world\n" in contents

    contents: Union[Union[str, bytes], Any] = logfile.read()
    assert b"[TRACE   ] hello there\n" in contents


@pytest.mark.skip(reason="Test doesn't work with nox, but works with a local run")
@pytest.mark.asyncio
async def test_async(hub, logfile, event_loop):
    hub.pop.loop.CURRENT_LOOP = event_loop
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=async",
            "--log-level=trace",
            f"--log-file={logfile.name}",
            "--log-handler-options",
            "mode=a",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    hub.log.trace("hello world")

    # Give time for the aiologger to flush
    await asyncio.sleep(0.5)

    contents = logfile.read()
    assert b"[TRACE   ] hello world\n" in contents


@pytest.mark.asyncio
async def test_async_rotating(hub, logfile, event_loop):
    hub.pop.loop.CURRENT_LOOP = event_loop
    with mock.patch(
        "sys.argv",
        [
            "my-app",
            "--log-plugin=rotating",
            "--log-level=trace",
            f"--log-file={logfile.name}",
            "--log-handler-options",
            "mode=a",
            "maxBytes=1",
            "backupCount=1",
        ],
    ):
        hub.config.integrate.load("pop_config", "pop_config")

    # Make sure we overflow and roll the log
    for _ in range(10):
        hub.log.trace("hello world")

    with open(f"{logfile.name}.1") as fh:
        contents = fh.read()

    assert "[TRACE   ] hello world\n" in contents
